import numpy as np
npa = np.array

class SoftmaxGroup:

class Softmax:
    """ Scaled sigmoid activation function"""
    def __init__(self, outMin, outMax):
        self.outMin = outMin
        self.outMax = outMax
        self.outScale = outMax - outMin

    def value(self, x):
        """:returns: scaled sigmoid"""

        return self.outMin + self.outScale * 1.0 / (1.0 + np.exp(-1 * x))

    def derivative(self, x):
        """:returns: scaled sigmoid's derivative"""
        logSigmoid = 1.0 / (1.0 + np.exp(-1 * x))
        return self.outScale * logSigmoid * (1.0 - logSigmoid)



def softmax(w, t=1.0):
    e = np.exp(npa(w) / t)
    dist = e / np.sum(e)
    return dist

if __name__ == '__main__':

    w = np.array([0.0, 0.0])
    print (softmax(w))

    w = np.array([-0.1, 0.2])
    print (softmax(w))

    w = np.array([0.9, -10])
    print (softmax(w))



"""Module with some basic neuron activation functions.
"""

import numpy as np
import matplotlib.pyplot as plt


class Identity:
    """Identity activation function"""
    def value(self, x):
        """:returns: x"""
        return x

    def derivative(self, x):
        """:returns: always 1.0"""
        return 1.0

# TODO: rename to sigmoid?
class LogSigmoid:
    """ Scaled sigmoid activation function"""
    def __init__(self, outMin, outMax):
        self.outMin = outMin
        self.outMax = outMax
        self.outScale = outMax - outMin

    def value(self, x):
        """:returns: scaled sigmoid"""

        return self.outMin + self.outScale * 1.0 / (1.0 + np.exp(-1 * x))

    def derivative(self, x):
        """:returns: scaled sigmoid's derivative"""
        logSigmoid = 1.0 / (1.0 + np.exp(-1 * x))
        return self.outScale * logSigmoid * (1.0 - logSigmoid)


def main():
    l = LogSigmoid(-10, 10)
    xx = np.linspace(-10, 10, 100)
    yy = [l.value(x) for x in xx]
    dd = [l.derivative(x) for x in xx]
    plt.plot(xx, yy)
    plt.plot(xx, dd)
    plt.show()


if __name__ == "__main__":
    main()



import graph_tool.all as gt
import numpy as np

from gtnn.network.activation import Identity


class Net(object):

    """This is a wrapper for the graph_tool.Graph object.
    One can register vertex/edge properties and use them in algorithm implementation.
    It also implements basic forward and backward operations.
    """

    def __init__(self,  nInput, nOutput, graph=gt.Graph()):
        self.g = gt.Graph(graph)
        self.nInput = nInput
        self.nOutput = nOutput
        self.properties = dict()
        self.biasProp = self.addVertexProperty("bias", "long double")
        self.valueProp = self.addVertexProperty("value", "long double")
        self.sumProp = self.addVertexProperty("sum", "long double")
        self.activation = self.addVertexProperty("activation", "python::object")
        self.errorProp = self.addVertexProperty("errorProp", "long double")
        self.weightProp = self.addEdgeProperty("weight", "long double")

        self.prepare()

    def addVertexProperty(self, name, typeName):
        self.properties[name] = self.g.new_vertex_property(typeName)
        return self.properties[name]

    def addEdgeProperty(self, name, typeName):
        self.properties[name] = self.g.new_edge_property(typeName)
        return self.properties[name]

    def prepare(self):
        """
        Computes ans saves the topological sort of the graph for future use.
        """
        self.order = np.array(gt.topological_sort(self.g)[::-1])
        for vIdx in self.order[:self.nInput]:
            self.activation[self.g.vertex(vIdx)] = Identity()

    def forward(self, inputVals=[]):
        g = self.g
        vp = self.valueProp
        bp = self.biasProp
        wp = self.weightProp
        sm = self.sumProp
        activation = self.activation

        for inpVal, vIdx in zip(inputVals, self.order[:self.nInput]):
            vp[g.vertex(vIdx)] = inpVal

        for vIdx in self.order[self.nInput:]:
            v = g.vertex(vIdx)
            inputs = np.array([vp[e.source()] for e in v.in_edges()])
            weights = np.array([wp[e] for e in v.in_edges()])

            sum = np.sum(inputs * weights) + bp[v]
            sm[v] = sum
            vp[v] = activation[v].value(sum)

        return np.array(
            [vp[g.vertex(vIdx)] for vIdx in self.order[-self.nOutput:]]
        )

    def backward(self, outputErr=[]):
        g = self.g
        ep = self.errorProp
        wp = self.weightProp
        sm = self.sumProp

        activation = self.activation

        for outErr, vIdx in zip(outputErr,
                                self.order[-self.nOutput:]):
            v = g.vertex(vIdx)
            ep[v] = outErr * activation[v].derivative(sm[v])

        # TODO: Change the borders
        for vIdx in reversed(self.order[:-self.nOutput]):
            v = g.vertex(vIdx)
            errors = np.array([ep[e.target()] for e in v.out_edges()])
            weights = np.array([wp[e] for e in v.out_edges()])
            ep[v] = np.sum(errors * weights) * activation[v].derivative(sm[v])

    def __str__(self):
        # TODO: Change output format
        ret = "Net: |V|=" + str(self.g.num_vertices())
        ret += " |E|=" + str(self.g.num_edges()) + "\n"

        for key, prop in self.properties.items():
            ret += key + ": " + str(prop.a) + "\n"
        return ret



import graph_tool.all as gt
import numpy as np


class LayeredNet(object):

    """This is a wrapper for the graph_tool.Graph object.
    One can register vertex/edge properties and use them in algorithm implementation.
    It also implements basic forward and backward operations.

    One can create a layered architecture with addLayer() method.
    Layers can overlap (can have some common neurons).
    The first layer is the input layer and the last one is the output layer.
    The output of the n-th layer is computed after the n-1 layer.

    Activation function is computed from the whole layer, so various non-trivial
    applications are possible. For example we create a Softmax layer only by
    introducing a suitable activation function..
    """

    def __init__(self,  nInput, nOutput, graph=gt.Graph()):
        self.g = gt.Graph(graph)
        self.properties = dict()
        self.biasProp = self.addVertexProperty("bias", "long double")
        self.valueProp = self.addVertexProperty("value", "long double")
        self.sumProp = self.addVertexProperty("sum", "long double")
        self.activation = self.addVertexProperty("activation", "python::object")
        self.errorProp = self.addVertexProperty("errorProp", "long double")
        self.weightProp = self.addEdgeProperty("weight", "long double")

        self.prepare()

    def addVertexProperty(self, name, typeName):
        self.properties[name] = self.g.new_vertex_property(typeName)
        return self.properties[name]

    def addEdgeProperty(self, name, typeName):
        self.properties[name] = self.g.new_edge_property(typeName)
        return self.properties[name]

    def addLayer(self):
        self.layers.append(self.g.new_vertex_property("bool"))
        return self.layers[-1]

    def prepare(self):
        """ Run after the layer structure is changed
        Creates helper GrahView objects"""
        self.layerGraphs = []
        for layerFilter in zip(self.layers[:-1], self.layers[1:]):
            self.layerGraphs.append(gt.GraphView(self.g,
                                                 vfilt=layerFilter))

    def forward(self, inputVals=[]):
        vp = self.valueProp
        bp = self.biasProp
        wp = self.weightProp
        sm = self.sumProp
        g = self.g
        activation = self.activation

        # Fill the input neurons
        # First layer is assumed to be the input layer.
        for inpVal, v in zip(inputVals,
                             self.layerGraphs[0].vertices()):
            vp[v] = inpVal

        for layerGraph in self.layerGraphs[1:]:
            # Compute the sum of inputs for every neuron in this layer
            for vLr in layerGraph.vertices():
                v = g.vertex(vLr)
                inputs = np.array([vp[e.source()] for e in v.in_edges()])
                weights = np.array([wp[e] for e in v.in_edges()])

                sm[v] = np.sum(inputs * weights) + bp[v]

            # NOTE: The whole layer has the same activation function.
            #       It could be a graph property
            a = activation[layerGraph.vertices().next()]
            activations = a.value([sm[vLr] for vLr in layerGraph.vertices()])

            # Fill the layer's values
            for v, act in zip(layerGraph.vertices(), activations):
                vp[v] = act

        return np.array([vp[v] for v in self.layerGraphs[-1].vertices()])

    def backward(self, outputErr=[]):
        g = self.g
        ep = self.errorProp
        wp = self.weightProp
        sm = self.sumProp
        activation = self.activation
        outLayer = self.layerGraphs[-1]

        # Compute the derivative on the whole layer:
        # NOTE: The whole layer has the same activation function.
        #       It could be a graph property
        ap = activation[outLayer.vertices().next()]
        derivatives = ap.derivative([sm[vLr] for vLr in outLayer.vertices()])

        # Fill the output neuron's errors
        for v, outErr, dv in zip(outLayer.vertices(),
                                 outputErr,
                                 derivatives):
            ep[v] = outErr * dv

        # Fill the errors in the other layers
        for layerGraph in reversed(self.layerGraphs[:-1]):
            for vLr in layerGraph.vertices():
                v = g.vertex(vLr)
                errors = np.array([ep[e.target()] for e in v.out_edges()])
                weights = np.array([wp[e] for e in v.out_edges()])
                ep[v] = np.sum(errors * weights)

            ap = activation[layerGraph.vertices().next()]
            derivatives = ap.derivative([sm[vLr] for vLr in layerGraph.vertices()])

            # Correct the errors with the derivative
            for vLr, dv in zip(layerGraph.vertices(),
                               derivatives):
                ep[vLr] *= dv

    def __str__(self):
        # TODO: Change output format
        ret = "Net: |V|=" + str(self.g.num_vertices())
        ret += " |E|=" + str(self.g.num_edges()) + "\n"

        for key, prop in self.properties.items():
            ret += key + ": " + str(prop.a) + "\n"
        return ret



import numpy as np

from gtnn.network.activation import LogSigmoid
from gtnn.network.net import Net


def mlp(sizes,
        weightGenerator=np.random.random,
        biasGenerator=np.random.random,
        activationFunction=LogSigmoid(-1, 1)):
    r"""Generate Multilayer Perceptron (MLP) and return as a Net

    :param sizes: Size of each layer
    :type sizes: list of integers
    :param functor weightGenerator: Functor for generating weights
    :param functor biasGenerator: Functor for generating bias values
    :param activationFunction: Activation function - the same for all neurons
    :type activationFunction: Activation function
    :returns: Net object with the requested architecture
    :rtype: :class:`~gtnn.network.Net`

    """
    n = Net(sizes[0], sizes[-1])
    layerId = n.addVertexProperty("layerId", "short")
    lastLayer = []
    presentLayer = []

    # Create all layers
    for layerIdx, size in enumerate(sizes):
        layerProp = n.addLayer()  # TODO: test layer
        for i in range(size):
            v = n.g.add_vertex()
            layerProp[v] = True
            layerId[v] = layerIdx
            n.activation[v] = activationFunction
            n.biasProp[v] = biasGenerator()

            presentLayer.append(v)
            for l in lastLayer:
                e = n.g.add_edge(l, v)
                n.weightProp[e] = weightGenerator()

        lastLayer = list(presentLayer)
        presentLayer = list()

    # Create a subgraph
    n.prepare()
    return n



import numpy as np


def train(net, inputs, outputs, numEpochs=100, learningRate=0.1):
    inputs = np.array(inputs)
    outputs = np.array(outputs)

    epochIdx = 0
    while True:
        for inp, out in zip(inputs, outputs):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Weights learning:
            for e in net.g.edges():
                net.weightProp[e] -= net.valueProp[e.source()] * \
                    learningRate * net.errorProp[e.target()]

            # Bias learning
            for v in net.g.vertices():
                net.biasProp[v] -= learningRate * net.errorProp[v]

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break


def miniBatchTrain(net,
                   inputs,
                   outputs,
                   numEpochs=100,
                   learningRate=0.1,
                   batchSize=10):
    inputs = np.array(inputs)
    outputs = np.array(outputs)
    epochIdx = 0

    sumWeightChangeProp = net.addEdgeProperty("sum weight change",
                                              "long double")

    sumBiasChangeProp = net.addVertexProperty("sum bias change",
                                              "long double")

    while True:
        for idx, (inp, out) in enumerate(zip(inputs, outputs)):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Summing up the changes in weights and bias
            for e in net.g.edges():
                sumWeightChangeProp[e] += net.valueProp[e.source()] * \
                    net.errorProp[e.target()]

            for v in net.g.vertices():
                sumBiasChangeProp[v] += net.errorProp[v]

            if idx % batchSize == 0:
                # Weights learning:
                for e in net.g.edges():
                    net.weightProp[e] -= learningRate * \
                        sumWeightChangeProp[e] / batchSize

                # Bias learning
                for v in net.g.vertices():
                    net.biasProp[v] -= learningRate * \
                        sumBiasChangeProp[v] / batchSize

                # Reset the changes
                sumWeightChangeProp.a = np.zeros(net.g.num_edges())
                sumBiasChangeProp.a = np.zeros(net.g.num_vertices())

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break


def batchTrain(net,
               inputs,
               outputs,
               numEpochs=100,
               learningRate=0.1):

    inputs = np.array(inputs)
    outputs = np.array(outputs)
    batchSize = len(inputs)

    sumWeightChangeProp = net.addEdgeProperty("sum weight change",
                                              "long double")

    sumBiasChangeProp = net.addVertexProperty("sum bias change",
                                              "long double")

    epochIdx = 0
    while True:
        for inp, out in zip(inputs, outputs):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Summing up the changes in weights and bias
            for e in net.g.edges():
                sumWeightChangeProp[e] += net.valueProp[e.source()] * \
                    net.errorProp[e.target()]

            for v in net.g.vertices():
                sumBiasChangeProp[v] += net.errorProp[v]

        # Weights learning:
        for e in net.g.edges():
            net.weightProp[e] -= learningRate * \
                sumWeightChangeProp[e] / batchSize

        # Bias learning
        for v in net.g.vertices():
            net.biasProp[v] -= learningRate * \
                sumBiasChangeProp[v] / batchSize

        # Reset the changes
        sumWeightChangeProp.a = np.zeros(net.g.num_edges())
        sumBiasChangeProp.a = np.zeros(net.g.num_vertices())

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break



import numpy as np

# TODO: Rename online


def train(net, inputs, outputs, numEpochs=100, learningRate=0.1, momentum=0.6):
    inputs = np.array(inputs)
    outputs = np.array(outputs)

    prevWeightChangeProp = net.addEdgeProperty("previous weight change",
                                               "long double")

    prevBiasChangeProp = net.addVertexProperty("previous bias change",
                                               "long double")
    epochIdx = 0
    while True:
        for inp, out in zip(inputs, outputs):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Weights learning:
            for e in net.g.edges():
                prevWeightChangeProp[e] = -net.valueProp[e.source()] * \
                    learningRate * net.errorProp[e.target()] + \
                    prevWeightChangeProp[e] * momentum

            net.weightProp.a += prevWeightChangeProp.a

            # Bias learning
            prevBiasChangeProp.a = -learningRate * net.errorProp.a + \
                momentum * prevBiasChangeProp.a

            net.biasProp.a += prevBiasChangeProp.a

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break


def miniBatchTrain(net,
                   inputs,
                   outputs,
                   numEpochs=100,
                   learningRate=0.1,
                   momentum=0.6,
                   batchSize=10):
    inputs = np.array(inputs)
    outputs = np.array(outputs)

    prevWeightChangeProp = net.addEdgeProperty("previous weight change",
                                               "long double")

    prevBiasChangeProp = net.addVertexProperty("previous bias change",
                                               "long double")

    sumWeightChangeProp = net.addEdgeProperty("sum weight change",
                                              "long double")

    sumBiasChangeProp = net.addVertexProperty("sum bias change",
                                              "long double")

    epochIdx = 0
    while True:
        for idx, (inp, out) in enumerate(zip(inputs, outputs)):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Summing up the changes in weights and bias
            for e in net.g.edges():
                sumWeightChangeProp[e] += net.valueProp[e.source()] * \
                    net.errorProp[e.target()]

            for v in net.g.vertices():
                sumBiasChangeProp[v] += net.errorProp[v]

            if idx % batchSize == 0:

                # Weights learning:
                for e in net.g.edges():
                    prevWeightChangeProp[e] = -learningRate * \
                        sumWeightChangeProp[e] / batchSize + \
                        prevWeightChangeProp[e] * momentum

                net.weightProp.a += prevWeightChangeProp.a

                # Bias learning
                prevBiasChangeProp.a = -learningRate * \
                    sumBiasChangeProp.a / batchSize + \
                    momentum * prevBiasChangeProp.a

                net.biasProp.a += prevBiasChangeProp.a

                # Reset the changes
                sumWeightChangeProp.a = np.zeros(net.g.num_edges())
                sumBiasChangeProp.a = np.zeros(net.g.num_vertices())

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break


def batchTrain(net,
               inputs,
               outputs,
               numEpochs=100,
               learningRate=0.1,
               momentum=0.6):
    inputs = np.array(inputs)
    outputs = np.array(outputs)

    batchSize = len(inputs)
    prevWeightChangeProp = net.addEdgeProperty("previous weight change",
                                               "long double")

    prevBiasChangeProp = net.addVertexProperty("previous bias change",
                                               "long double")

    sumWeightChangeProp = net.addEdgeProperty("sum weight change",
                                              "long double")

    sumBiasChangeProp = net.addVertexProperty("sum bias change",
                                              "long double")

    epochIdx = 0
    while True:
        for idx, (inp, out) in enumerate(zip(inputs, outputs)):
            netOut = net.forward(inp)
            err = netOut - out
            net.backward(err)

            # Summing up the changes in weights and bias
            for e in net.g.edges():
                sumWeightChangeProp[e] += net.valueProp[e.source()] * \
                    net.errorProp[e.target()]

            for v in net.g.vertices():
                sumBiasChangeProp[v] += net.errorProp[v]

        # Weights learning:
        for e in net.g.edges():
            prevWeightChangeProp[e] = -learningRate * \
                sumWeightChangeProp[e] / batchSize + \
                prevWeightChangeProp[e] * momentum

        net.weightProp.a += prevWeightChangeProp.a

        # Bias learning
        prevBiasChangeProp.a = -learningRate * \
            sumBiasChangeProp.a / batchSize + \
            momentum * prevBiasChangeProp.a

        net.biasProp.a += prevBiasChangeProp.a

        # Reset the changes
        sumWeightChangeProp.a = np.zeros(net.g.num_edges())
        sumBiasChangeProp.a = np.zeros(net.g.num_vertices())

        # End condition
        epochIdx += 1
        if epochIdx > numEpochs:
            break



from distutils.core import setup

setup(
    name='graph-tool-nn',
    version='1.1.0',
    packages=['gtnn', 'gtnn.generators', 'gtnn.learn', 'gtnn.network'],
    license='GNU General Public License v3 (GPLv3)',
    long_description=r'''Graph-tool is a great open source tool for creating, using and analyzing graphs. It's a python
library with C++ bindings, uses boost::graph under the hood and seems to be pretty fast
(http://graph-tool.skewed.de/).

Graph-tool Neural Networks (gtnn) is an implementation of ANN on top of graph-tool. It makes
researching neural networks nice&easy. You can create custom nets, train, analyze and plot them.

''',

    author="Jan Olszak",
    author_email="janekolszak@gmail.com",
    description="Artificial neural networks implementation using graph-tool",
    keywords="graph-tool neural networks",
    url="http://github.com/janekolszak/gtnn",

    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
)
